# Copyright (c) 2018, Neil Booth
#
# All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

'''RPC message framing in a byte stream.'''

__all__ = ('FramerBase', 'NewlineFramer')

from itertools import chain
from collections import namedtuple
import struct


class FramerBase(object):
    '''Abstract base class for a framer.

    A framer breaks an incoming byte stream into protocol messages,
    buffering if necesary.  It also frames outgoing messages into
    a byte stream.
    '''

    def frame(self, messages):
        '''Return bytes formed by framing each message in messages,
        an iterable, and concatenating the result.
        '''
        raise NotImplementedError

    def messages(self, data):
        '''A generator that yields messages for the caller to process.

        Raises a MemoryError Exception if the internal buffer
        overflows, so converting to a list before processing risks
        message loss.
        '''
        raise NotImplementedError


class NewlineFramer(FramerBase):
    '''A framer for a protocol where messages are separated by newlines.'''

    def __init__(self, max_size=250 * 4000):
        '''max_size - an anti-DoS measure.  If, after processing an incoming
        message, buffered data would exceed max_size bytes, that
        buffered data is dropped entirely and the framer waits for a
        newline character to re-synchronize the stream.
        '''
        # The default max_size value is motivated by JSONRPC, where a
        # normal request will be 250 bytes or less, and a reasonable
        # batch may contain 4000 requests.
        self.max_size = max_size
        self.parts = []
        self.synchronizing = False

    def frame(self, messages):
        return b''.join(chain.from_iterable((msg, b'\n') for msg in messages))

    def messages(self, data):
        assert isinstance(data, (bytes, bytearray))
        parts = self.parts
        while True:
            npos = data.find(ord('\n'))
            if npos == -1:
                parts.append(data)
                break
            tail, data = data[:npos], data[npos + 1:]
            if self.synchronizing:
                self.synchronizing = False
            else:
                parts.append(tail)
                yield b''.join(parts)
                parts.clear()

        # Ignore over-sized messages; re-synchronize
        buffer_size = sum(len(part) for part in parts)
        if buffer_size > self.max_size:
            parts.clear()
            if not self.synchronizing:
                self.synchronizing = True
                raise MemoryError(f'dropping message over {self.max_size} '
                                  ' and re-synchronizing')



from .curio import *
from .framing import *
from .jsonrpc import *
from .rpc import *
from .socks import *
from .session import *
from .util import *

_version = (0, 5, 8)
_version_str = '.'.join(str(part) for part in _version)

__all__ = (curio.__all__ +
           framing.__all__ +
           jsonrpc.__all__ +
           rpc.__all__ +
           socks.__all__ +
           session.__all__ +
           util.__all__)



# Copyright (c) 2018, Neil Booth
#
# All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


__all__ = ('ClientSession', 'ServerSession', 'Server', 'ConnectionError')


import asyncio
import logging
import ssl
import time

from .framing import NewlineFramer
from .jsonrpc import JSONRPCv2
from .rpc import (RPCProcessor, RPCRequest, RPCRequestOut,
                  RPCBatchOut, RPCHelperBase)
from .util import TaskSet, Concurrency


class ConnectionError(RuntimeError):
    pass


class SessionBase(asyncio.Protocol, RPCHelperBase):

    concurrency_recalc_interval = 15
    max_errors = 10

    def __init__(self, rpc_protocol=None, framer=None, loop=None):
        self.loop = loop or asyncio.get_event_loop()
        self.logger = logging.getLogger(self.__class__.__name__)
        rpc_protocol = rpc_protocol or self.default_rpc_protocol()
        self.rpc = RPCProcessor(rpc_protocol, self)
        self.framer = framer or self.default_framer()
        self.transport = None
        # Concurrency
        self.max_concurrent = 6
        # Set when a connection is made
        self._address = None
        self._proxy_address = None
        # For logger.debug messsages
        self.verbosity = 0
        # Pausing sends when socket is full
        self.paused = False
        self.paused_messages = []
        # Close once the send queue is exhausted.  An unfortunate hack
        # necessitated by asyncio's non-blocking socket writes
        self.close_after_send = False
        # Statistics.  The RPC object also keeps its own statistics.
        self.start_time = time.time()
        self.send_count = 0
        self.send_size = 0
        self.last_send = self.start_time
        self.recv_count = 0
        self.recv_size = 0
        self.last_recv = self.start_time
        # Bandwidth usage per hour before throttling starts
        self.bw_limit = 2000000
        self.bw_time = self.start_time
        self.bw_charge = 0
        # Asyncronous tasks
        self.tasks = TaskSet()
        self.concurrency = Concurrency(self.max_concurrent)

    def semaphore(self):
        return self.concurrency.semaphore

    def using_bandwidth(self, size):
        '''Called when sending or receiving size bytes.'''
        self.bw_charge += size

    async def _concurrency_loop(self):
        '''Reclaculate concurrency setting at regular intervals.'''
        while True:
            await self._update_concurrency()
            await asyncio.sleep(self.concurrency_recalc_interval)

    async def _update_concurrency(self):
        now = time.time()
        # Reduce the recorded usage in proportion to the elapsed time
        refund = (now - self.bw_time) * (self.bw_limit / 3600)
        self.bw_charge = max(0, self.bw_charge - int(refund))
        self.bw_time = now
        # Reduce concurrency allocation by 1 for each whole bw_limit used
        throttle = int(self.bw_charge / self.bw_limit)
        target = max(1, self.max_concurrent - throttle)
        current = self.concurrency.max_concurrent
        if target != current:
            self.logger.info(f'changing task concurrency from {current} '
                             f'to {target}')
            await self.concurrency.set_max_concurrent(target)

    def data_received(self, framed_message):
        '''Called by asyncio when a message comes in.'''
        if self.verbosity >= 4:
            self.logger.debug(f'Received framed message {framed_message}')
        self.recv_size += len(framed_message)
        self.using_bandwidth(len(framed_message))

        count = 0
        try:
            for message in self.framer.messages(framed_message):
                count += 1
                self.rpc.message_received(message)
        except MemoryError as e:
            self.logger.warning(str(e))

        if count:
            self.recv_count += count
            self.last_recv = time.time()

    def send_message(self, message):
        '''Send a message over the connection. It is framed before sending.'''
        framed_message = self.framer.frame((message, ))
        self.send_framed_message(framed_message)

    def send_framed_message(self, framed_message):
        if self.is_closing():
            return
        if self.paused:
            self.paused_messages.append(framed_message)
        else:
            self.send_size += len(framed_message)
            self.using_bandwidth(len(framed_message))
            self.send_count += 1
            self.last_send = time.time()
            if self.verbosity >= 4:
                self.logger.debug(f'Sending framed message {framed_message}')
            self.transport.write(framed_message)
            if self.close_after_send or self.rpc.errors >= self.max_errors:
                self.close()

    def pause_writing(self):
        '''Transport calls when the send buffer is full.'''
        self.logger.info('pausing whilst socket drains')
        self.paused = True
        self.transport.pause_reading()

    def resume_writing(self):
        '''Transport calls when the send buffer has room.'''
        self.logger.info('resuming processing')
        self.paused = False
        self.transport.resume_reading()
        self.send_framed_message(b''.join(self.paused_messages))
        self.paused_messages.clear()

    def peer_address(self):
        '''Returns the peer's address (Python networking address), or None if
        no connection or an error.

        This is the result of socket.getpeername() when the connection
        was made.
        '''
        return self._address

    def peer_address_str(self):
        '''Returns the peer's IP address and port as a human-readable
        string.'''
        if not self._address:
            return 'unknown'
        ip_addr_str, port = self._address[:2]
        if ':' in ip_addr_str:
            return f'[{ip_addr_str}]:{port}'
        else:
            return f'{ip_addr_str}:{port}'

    # External API
    def default_rpc_protocol(self):
        '''Return a default rpc helper if the user provides none.'''
        return JSONRPCv2

    def default_framer(self):
        '''Return a default framer.'''
        return NewlineFramer()

    def create_task(self, coro):
        return self.tasks.create_task(coro)

    def connection_made(self, transport):
        '''Called by asyncio when a connection is established.

        Derived classes overriding this method must call this first.'''
        self.transport = transport
        # This would throw if called on a closed SSL transport.  Fixed
        # in asyncio in Python 3.6.1 and 3.5.4
        peer_address = transport.get_extra_info('peername')
        # If the Socks proxy was used then _address is already set to
        # the remote address
        if self._address:
            self._proxy_address = peer_address
        else:
            self._address = peer_address
        self.tasks.create_task(self._concurrency_loop())

    def connection_lost(self, exc):
        '''Called by asyncio when the connection closes.

        Tear down things done in connection_made.'''
        self._address = None
        self.transport = None
        self.tasks.cancel_all()
        for request in self.all_requests():
            request.set_exception(ConnectionError(
                'connection lost before request completed'))

    # App layer
    async def wait_closed(self):
        '''Returns when all pending tasks and requests have completed.'''
        await self.tasks.wait()
        await asyncio.gather(*self.all_requests(), return_exceptions=True)

    def is_closing(self):
        '''Return True if the connection is closing.'''
        return not self.transport or self.transport.is_closing()

    def close(self):
        '''Close the connection.'''
        if self.transport:
            self.transport.close()

    def abort(self):
        '''Cut the connection abruptly.'''
        if self.transport:
            self.transport.abort()

    def send_request(self, method, args=None, on_done=None, *, timeout=None):
        '''Send an RPC request over the network.'''
        request = RPCRequestOut(method, args, on_done, loop=self.loop)
        self.rpc.send_request(request)
        if timeout is not None:
            self.set_timeout(request, timeout)
        return request

    def send_notification(self, method, args=None):
        '''Send an RPC notification over the network.'''
        request = RPCRequest(method, args, None)
        self.rpc.send_request(request)
        return request

    def new_batch(self):
        return RPCBatchOut(loop=self.loop)

    def send_batch(self, batch, on_done=None, *, timeout=None):
        self.rpc.send_batch(batch, on_done)
        if timeout is not None:
            self.set_timeout(batch, timeout)
        return batch

    def set_timeout(self, request, delay):
        '''Cause a request (or batch request) to time-out after delay
        seconds (a float).'''
        if request not in self.rpc.all_requests():
            raise RuntimeError(f'cannot set a timeout on {request} - it does '
                               'not belong to this session')

        def on_timeout():
            msg = f'{request} timed out after {delay}s'
            request.set_exception(asyncio.TimeoutError(msg))

        def on_done(request):
            handle.cancel()

        handle = self.loop.call_later(delay, on_timeout)
        request.add_done_callback(on_done)

    def all_requests(self):
        '''Returns a list of all requests that have not yet completed.

        If a batch requests is outstanding, it is returned and not the
        individual requests it is comprised of.
        '''
        return self.rpc.all_requests()


class ClientSession(SessionBase):
    '''A client session.

    To initiate a connection to the remote server call
    create_connection().  Each successful call should have a
    corresponding call to close().

    Alternatively if used in a with statement, the connection is made
    on entry to the block, and closed on exit from the block.
    '''

    def __init__(self, host=None, port=None, *, rpc_protocol=None, framer=None,
                 loop=None, proxy=None, **kwargs):
        super().__init__(rpc_protocol, framer, loop)
        self.host = host
        self.port = port
        self.kwargs = kwargs
        self.proxy = proxy

    async def create_connection(self):
        '''Initiate a connection.'''
        def self_func():
            return self
        if self.proxy:
            return await self.proxy.create_connection(
                self_func, self.host, self.port, loop=self.loop, **self.kwargs)

        return await self.loop.create_connection(
            self_func, self.host, self.port, **self.kwargs)

    async def __aenter__(self):
        await self.create_connection()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        self.close()


class ServerSession(SessionBase):
    '''A server session - created by a listening Server for each incoming
    connection.'''


class Server(object):
    '''A simple wrapper around an asyncio.Server object.'''

    def __init__(self, session_factory, host=None, port=None, *,
                 loop=None, **kwargs):
        self.host = host
        self.port = port
        self.loop = loop or asyncio.get_event_loop()
        self.server = None
        self._session_factory = session_factory
        self._kwargs = kwargs
        self._sessions = set()

    async def listen(self):
        self.server = await self.loop.create_server(
            self._session_factory, self.host, self.port, **self._kwargs)

    async def wait_closed(self):
        if self.server:
            await self.server.wait_closed()
            self.server = None

    def close(self):
        '''Close the listening socket.  This does not close any ServerSession
        objects created to handle incoming connections.
        '''
        if self.server:
            self.server.close()



# Copyright (c) 2018, Neil Booth
#
# All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

'''RPC message processing, independent of transport and RPC protocol.'''

__all__ = ('RPCError', 'RPCProtocolBase')

from asyncio import Future, CancelledError
from functools import partial
import itertools
import logging

from .util import is_async_call, signature_info


class RPCRequest(object):
    '''An RPC request or notification that has been received, or an
    outgoing notification.

    Outgoing requests are represented by RPCRequestOut.
    '''

    def __init__(self, method, args, request_id):
        self.method = method
        if args is None:
            self.args = []
        else:
            if not isinstance(args, (list, tuple, dict)):
                raise ValueError('request args must be a list or dictionary')
            self.args = args
        self.request_id = request_id
        # Ill-formed requests for which the protocol couldn't
        # determine a meaning pass an RPCError as their method
        if isinstance(method, RPCError):
            method.request_id = request_id

    def __repr__(self):
        return (f'RPCRequest({self.method!r}, {self.args!r}, '
                f'{self.request_id!r})')

    def is_notification(self):
        return self.request_id is None


class RPCRequestOut(RPCRequest, Future):
    '''Represents an outgoing RPC request which expects a response.

    You can specify a callback to call when a response arrives; it is
    passed the request object.  The result can be retrieved with its
    result() method.

    The request can be await-ed on and/or it can be given a handler to
    call on completion.
    '''

    id_counter = itertools.count()

    def __init__(self, method, args, on_done, *, loop=None):
        '''Initialize a request using the next unique request ID.

        on_done - can be None
        '''
        RPCRequest.__init__(self, method, args, next(self.id_counter))
        Future.__init__(self, loop=loop)
        if on_done:
            self.add_done_callback(on_done)


class RPCResponse(object):
    '''An RPC response, incoming or outgoing.

    An error is indicated by a result that is an RPCError.
    '''

    def __init__(self, result, request_id):
        # result is an RPCError object if an error was returned
        self.result = result
        self.request_id = request_id
        if isinstance(result, RPCError):
            result.request_id = request_id

    def __repr__(self):
        return f'RPCResponse({self.result!r}, {self.request_id!r})'


class RPCError(Exception):
    '''An RPC error.

    When an RPC error response is received, an object of this type is
    embedded in the RPCResponse object as its "result" and passed to
    the user-defined response handler.

    When protocol.message_to_item() parses an incoming request (or
    batch item), if it is ill-formed (for example, it cannot be
    parsed, or the method name is not a string),
    protocol.message_to_item() should return it embedded in an
    RPCRequest object as its "method".  When the request is processed
    the framework will embed it in a RPCResponse object to send over
    the network.

    A request handler can raise it to cause the framework to send an
    error response.
    '''

    def __init__(self, code, message, request_id=None):
        super().__init__(message, code)
        self.message = message
        self.code = code
        self.request_id = request_id

    def __repr__(self):
        if self.request_id is None:
            return f'RPCError({self.code:d}, {self.message!r})'
        else:
            return (f'RPCError({self.code:d}, {self.message!r}, '
                    f'{self.request_id!r})')

    __str__ = __repr__


class RPCBatch(object):
    '''An incoming or outgoing RPC response batch, or an incoming RPC
    request batch.
    '''

    def __init__(self, items):
        self.items = items
        assert isinstance(items, list)
        assert items
        assert (all(isinstance(item, RPCRequest) for item in items) or
                all(isinstance(item, RPCResponse) for item in items))

    def requests(self):
        '''An iterable of the batch items that are not notifications.

        For a response batch simply returns everything.'''
        for item in self.items:
            if item.request_id is not None:
                yield item

    def request_ids(self):
        '''Return a frozenset of all request IDs in the batch, ignoring
        notifications.
        '''
        return frozenset(item.request_id for item in self.requests())

    def is_request_batch(self):
        return isinstance(self.items[0], RPCRequest)

    def __len__(self):
        return len(self.items)

    def __iter__(self):
        return iter(self.items)

    def __repr__(self):
        return f'RPCBatch({self.items!r})'


class RPCBatchOut(RPCBatch, Future):
    '''Represents an outgoing RPC batch request.

    You can specify a callback to call when all requests in the batch
    have completed; it is passed the batch object.  The batch object
    does not have a meaningful result.

    The batch can be await-ed on and/or it can be given a handler to
    call on completion.  This is also true for its member requests.
    '''
    def __init__(self, *, loop=None):
        '''Create an empty outgoing batch request.  Members can be
        added with add_request and add_notification.'''
        # We don't call RPCBatch.__init__()
        self.items = []
        Future.__init__(self, loop=loop)

    def _cancel_requests(self):
        for request in self.requests():
            request.cancel()

    def cancel(self):
        super().cancel()
        self._cancel_requests()

    def set_result(self, result):
        super().set_result(result)
        self._cancel_requests()

    def set_exception(self, exception):
        super().set_exception(exception)
        self._cancel_requests()

    def add_request(self, method, args=None, on_done=None):
        request = RPCRequestOut(method, args, on_done)
        self.items.append(request)
        return request

    def add_notification(self, method, args=None):
        request = RPCRequest(method, args, None)
        self.items.append(request)
        return request


class RPCHelperBase(object):
    '''Abstract base class of an object that handles RPC requests, job
    queueing and message sending for RPCProcessor.'''

    # The maximum response size.  Responses above this size send an
    # error response instead.
    max_response_size = 1000000

    def send_message(self, message):
        '''Called when there is a message to send over the network.  The
        message is unframed.  If a request was cancelled, send_message
        is called with an empty message, which should be ignored.

        The derived class may want to queue several messages and send
        them as a batch, or delay / throttle the sends in some way.
        '''
        raise NotImplementedError

    def create_task(self, coro):
        '''Create a task to run the given coroutine.  Return the task.'''
        raise NotImplementedError

    def semaphore(self):
        '''Return a semaphore to limit concurrency.'''
        raise NotImplementedError

    def notification_handler(self, method):
        '''Return the handler for the given notification.

        The handler can be synchronous or asynchronous.  When called
        the return value is ignored.
        '''
        return None

    def request_handler(self, method):
        '''Return the handler for the given request method.

        The handler can be synchronous or asynchronous.  The return value
        is sent as an RPC response.'''
        return None


class RPCProtocolBase(object):

    @classmethod
    def request_message(cls, item):
        '''Convert an RPCRequest item to a message.'''
        raise NotImplementedError

    @classmethod
    def response_message(cls, item):
        '''Convert an RPCResponse item to a message.'''
        raise NotImplementedError

    @classmethod
    def error_message(cls, item):
        '''Convert an RPCError item to a message.'''
        raise NotImplementedError

    @classmethod
    def batch_message(cls, item):
        '''Convert an RPCBatch item (a request batch) to a message.'''
        raise NotImplementedError

    @classmethod
    def batch_message_from_parts(cls, messages):
        '''Convert messages, one per batch item, into a batch message.  At
        least one message must be passed.
        '''
        raise NotImplementedError

    @classmethod
    def internal_error(cls, request_id):
        '''Return an internal error RPCError object.'''
        return RPCError(cls.INTERNAL_ERROR,
                        'internal error processing request', request_id)

    @classmethod
    def args_error(cls, message):
        '''Return an invalid args RPCError object.'''
        return RPCError(cls.INVALID_ARGS, message, None)

    @classmethod
    def invalid_request(cls, message, request_id=None):
        '''Return an invalid request RPCError object.'''
        return RPCError(cls.INVALID_REQUEST, message, request_id)

    @classmethod
    def method_not_found(cls, message):
        '''Return a method-not-found RPCError object.'''
        return RPCError(cls.METHOD_NOT_FOUND, message, None)


class RPCProcessor(object):
    '''Handles RPC message processing.

    Coordinates the processing of incoming and outgoing RPC requests,
    responses and notifications.
    '''

    def __init__(self, protocol, helper):
        self.protocol = protocol
        self.helper = helper
        self.logger = logging.getLogger(self.__class__.__name__)
        # Sent requests and batch requests awaiting a response.  For an
        # RPCRequestOut object the key is its request ID; for a batch
        # it is its frozenset of request IDs
        self.requests = {}
        # Statistics
        # Count of requests sent and received (a batch of 10 counts as 10)
        # Responses are not tracked
        self.sent_count = 0
        self.recv_count = 0
        # Number of requests or notifications that raised an RPCError
        # during processing because the request was invalid in some way.
        self.errors = 0
        # This includes internal errors, which are exceptions other than
        # RPCError raised processing requests.
        self.internal_errors = 0
        # Incoming unprocessed request count
        self.pending_requests = 0

    def _error_message(self, request, error, force=False):
        '''Return an error message when an exception (error) is raised.'''
        # Batches don't have a request ID
        request_id = getattr(request, 'request_id', None)
        if isinstance(error, RPCError):
            self.errors += 1
            error.request_id = request_id
            self.logger.debug('error processing request: %s %s',
                              repr(error), repr(request))
            # Notifications don't get a response, unless ill-formed
            if request_id is None and not force:
                return None
        else:
            self.internal_errors += 1
            self.logger.exception('exception raised processing request: %s',
                                  repr(request))
            if request_id is None:
                return None
            error = self.protocol.internal_error(request_id)

        return self.protocol.error_message(error)

    def _response_message(self, request, task):
        '''Return a response message to a request, or None if no message
        should be sent.  Log errors.

        Notifications return a response only if ill-formed.'''
        try:
            result = task.result()
            if request.request_id is None:
                return None
            response = RPCResponse(result, request.request_id)
            return self.protocol.response_message(response)
        except CancelledError as error:
            return None
        except Exception as error:
            return self._error_message(request, error,
                                       isinstance(request.method, RPCError))

    def _send_response(self, response, request):
        if len(response) > self.helper.max_response_size > 0:
            msg = f'response too large (at least {len(response)} bytes)'
            error = self.protocol.invalid_request(msg)
            response = self._error_message(request, error, True)
        self.helper.send_message(response)

    def _handle_request(self, request, task):
        response = self._response_message(request, task)
        if response:
            self._send_response(response, request)

    def _rpc_call(self, request):
        '''Return a partial function call that calls the RPC function
        to handle the request with the appropriate arguments.

        If the request is bad an RPCError is raised.  Any exceptions
        raised when determining the handler function are passed on.
        '''
        # Raise ill-formed requests here so that they are logged
        method = request.method
        if isinstance(method, RPCError):
            raise method

        # Let through any exceptions raised when determining the handler
        request_id = request.request_id
        if request_id is None:
            handler = self.helper.notification_handler(method)
        else:
            handler = self.helper.request_handler(method)
        if not handler:
            raise self.protocol.method_not_found(f'unknown method "{method}"')

        # We must test for too few and too many arguments.  How
        # depends on whether the arguments were passed as a list or as
        # a dictionary.
        info = signature_info(handler)
        args = request.args
        if isinstance(args, list):
            if len(args) < info.min_args:
                s = '' if len(args) == 1 else 's'
                raise self.protocol.args_error(
                    f'{len(args)} argument{s} passed to method '
                    f'"{method}" but it requires {info.min_args}')
            if info.max_args is not None and len(args) > info.max_args:
                s = '' if len(args) == 1 else 's'
                raise self.protocol.args_error(
                    f'{len(args)} argument{s} passed to method '
                    f'{method} taking at most {info.max_args}')
            return partial(handler, *args)

        # Arguments passed by name
        if info.other_names is None:
            raise self.protocol.args_error(f'method "{method}" cannot '
                                           f'be called with named arguments')

        missing = set(info.required_names).difference(args)
        if missing:
            s = '' if len(missing) == 1 else 's'
            missing = ', '.join(sorted(f'"{name}"' for name in missing))
            raise self.protocol.args_error(f'method "{method}" requires '
                                           f'parameter{s} {missing}')

        if info.other_names is not any:
            excess = set(args).difference(info.required_names)
            excess = excess.difference(info.other_names)
            if excess:
                s = '' if len(excess) == 1 else 's'
                excess = ', '.join(sorted(f'"{name}"' for name in excess))
                raise self.protocol.args_error(f'method "{method}" does not '
                                               f'take parameter{s} {excess}')
        return partial(handler, **args)

    async def _process_request(self, request):
        '''Process a request or notification and return its result.

        Raises RPCError to ill-formed or erroneous requests.'''
        self.recv_count += 1
        self.pending_requests += 1
        try:
            async with self.helper.semaphore():
                rpc_call = self._rpc_call(request)
                if is_async_call(rpc_call):
                    return await rpc_call()
                else:
                    return rpc_call()
        finally:
            self.pending_requests -= 1

    def _process_request_batch(self, batch):
        '''Process a request batch.  Create a task for each sub-request, and
        collect the results.  Send a response, if not empty, once all
        results have come in.
        '''
        def send_batch_response():
            response = self.protocol.batch_message_from_parts(parts)
            self._send_response(response, batch)

        def fail_batch():
            nonlocal failed
            failed = True
            for task in tasks:
                task.cancel()

        def collect_response(request, task):
            nonlocal failed, response_size
            # Collect the response even if we don't use it to avoid
            # diagnostics about uncomsumed exceptions
            response = self._response_message(request, task)
            if response:
                parts.append(response)
                response_size += len(response)
                if response_size > self.helper.max_response_size > 0:
                    send_batch_response()
                    fail_batch()
            tasks.remove(task)
            # If any task was cancelled fail the batch
            if task.cancelled() and not failed:
                fail_batch()
            if not tasks and parts and not failed:
                send_batch_response()

        parts = []
        failed = False
        response_size = 0
        tasks = set()
        for request in batch:
            task = self.helper.create_task(self._process_request(request))
            task.add_done_callback(partial(collect_response, request))
            tasks.add(task)

    def _handle_request_response(self, request, response):
        if isinstance(response.result, RPCError):
            self.logger.debug('request returned errror: %s %s',
                              repr(request), repr(response.result))
            request.set_exception(response.result)
        else:
            request.set_result(response.result)

    def _process_response(self, response):
        request_id = response.request_id
        if request_id is None:
            self.logger.debug('missing id: %s', repr(response))
            return

        request = self.requests.pop(request_id, None)
        if request:
            self._handle_request_response(request, response)
        else:
            self.logger.debug('response to unsent or forced request: %s',
                              repr(response))

    def _process_response_batch(self, batch):
        request_ids = batch.request_ids()
        batch_request = self.requests.pop(request_ids, None)
        if batch_request:
            requests_by_id = {item.request_id: item
                              for item in batch_request.requests()}
            for response in batch:
                request_id = response.request_id
                if request_id is None:
                    self.logger.debug('batch response missing id: %s',
                                      repr(response))
                else:
                    request = requests_by_id[request_id]
                    self._handle_request_response(request, response)
        else:
            self.logger.debug('response to unsent or forced '
                              'batch request: %s', repr(batch))

    # External API - methods for use by a session layer
    def message_received(self, message):
        '''Analyse an incoming message and queue it for processing.

        Any response will be sent to send_msessage.  This can happen
        before or after this function returns.
        '''
        item = self.protocol.message_to_item(message)
        if isinstance(item, RPCRequest):
            task = self.helper.create_task(self._process_request(item))
            task.add_done_callback(partial(self._handle_request, item))
        elif isinstance(item, RPCResponse):
            self._process_response(item)
        elif isinstance(item, RPCBatch):
            if item.is_request_batch():
                self._process_request_batch(item)
            else:
                self._process_response_batch(item)
        else:
            # Protocol auto-detection
            assert isinstance(item(), RPCProtocolBase)
            self.protocol = item
            self.message_received(message)

    def send_request(self, request):
        '''Send a request.

        If it is not a notification record the request ID so that an
        incoming response can be handled.
        '''
        self.sent_count += 1
        if isinstance(request, RPCRequestOut):
            def request_done(request):
                self.requests.pop(request.request_id, None)

            self.requests[request.request_id] = request
            request.add_done_callback(request_done)
        self.helper.send_message(self.protocol.request_message(request))

    def send_batch(self, batch, on_done=None):
        '''Send a batch request.

        Unless it is all notifications, record the request IDs of the
        batch memebers so that an incoming batch response can be
        handled.
        '''
        if not batch:
            raise RuntimeError('request batch cannot be empty')

        self.sent_count += len(batch)
        if on_done:
            batch.add_done_callback(on_done)

        request_ids = batch.request_ids()
        if request_ids:
            def request_done(request):
                nonlocal remaining
                remaining -= 1
                if not remaining and not batch.done():
                    batch.set_result(False)

            def batch_done(batch):
                self.requests.pop(request_ids, None)

            remaining = len(request_ids)
            self.requests[request_ids] = batch
            for request in batch.requests():
                request.add_done_callback(request_done)
            batch.add_done_callback(batch_done)
        else:
            batch.set_result(False)

        self.helper.send_message(self.protocol.batch_message(batch))

    def all_requests(self):
        '''Returns a list of all requests that have not yet completed.

        If a batch requests is outstanding, it is returned and not the
        individual requests it is comprised of.
        '''
        return list(self.requests.values())



# Copyright (c) 2018, Neil Booth
#
# All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

'''Classes for JSONRPC versions 1.0 and 2.0, and a loose interpretation.'''

__all__ = ('JSONRPC', 'JSONRPCv1', 'JSONRPCv2', 'JSONRPCLoose',
           'JSONRPCAutoDetect')

import json
import logging
import numbers
import traceback

from .rpc import RPCError, RPCBatch, RPCRequest, RPCResponse, RPCProtocolBase


class JSONRPC(RPCProtocolBase):
    '''Abstract base class that interprets and constructs JSON RPC messages.'''

    # Error codes.  See http://www.jsonrpc.org/specification
    PARSE_ERROR = -32700
    INVALID_REQUEST = -32600
    METHOD_NOT_FOUND = -32601
    INVALID_ARGS = -32602
    INTERNAL_ERROR = -32603
    # Codes specific to this library
    ERROR_CODE_UNAVAILABLE = -100

    # Can be overridden by client application
    logger = logging.getLogger('JSONRPC')

    # Can be overridden by derived classes
    allow_batches = True

    @classmethod
    def _message_id(cls, message, require_id):
        '''Validate the message is a dictionary and return its ID.

        Raise an error if the message is invalid or the ID is of an
        invalid type.  If it has no ID, raise an error if require_id
        is True, otherwise return None.
        '''
        raise NotImplementedError

    @classmethod
    def _validate_message(cls, message):
        '''Validate other parts of the message other than those
        done in _message_id.'''
        pass

    @classmethod
    def _request_args(cls, request):
        '''Validate the existence and type of the arguments passed
        in the request dictionary.'''
        raise NotImplementedError

    @classmethod
    def _process_batch(cls, batch):
        if not batch:
            return RPCRequest(cls.invalid_request('empty batches are invalid'),
                              None, None)

        if any(isinstance(payload, dict) and 'method' in payload
               for payload in batch):
            processor = cls._process_request
        else:
            processor = cls._process_response

        return RPCBatch([processor(payload) for payload in batch])

    @classmethod
    def _process_request(cls, payload):
        request_id = None
        try:
            request_id = cls._message_id(payload, False)
            cls._validate_message(payload)
            method = payload.get('method')
            if not isinstance(method, str):
                raise RPCError(cls.METHOD_NOT_FOUND,
                               'request method must be a string')
            return RPCRequest(method, cls._request_args(payload), request_id)
        except RPCError as error:
            return RPCRequest(error, None, request_id)
        except Exception:
            cls.logger.exception(f'error processing request {payload!r}')
            return RPCRequest(cls.internal_error(request_id), None,
                              request_id)

    @classmethod
    def _process_response(cls, payload):
        request_id = None
        try:
            request_id = cls._message_id(payload, True)
            cls._validate_message(payload)
            return RPCResponse(cls.response_result(payload), request_id)
        except RPCError as error:
            return RPCResponse(error, request_id)
        except Exception:
            cls.logger.exception(f'error processing response {payload!r}')
            return RPCResponse(cls.internal_error(request_id),
                               request_id)

    @classmethod
    def _message_to_payload(cls, message):
        '''Returns a Python object or an RPCError.'''
        try:
            return json.loads(message.decode())
        except UnicodeDecodeError as e:
            return RPCError(cls.PARSE_ERROR,
                            f'messages must be encoded in UTF-8: {e}')
        except json.JSONDecodeError as e:
            return RPCError(cls.PARSE_ERROR, f'cannot decode JSON: {e}')

    #
    # External API
    #
    @classmethod
    def message_to_item(cls, message):
        '''Convert a binary message to an RPCRequest, RPCResponse
        or RPCBatch object and return it.

        If the message is ill-formed an RPCRequest object with an RPCError
        as its method will be returned.'''
        payload = cls._message_to_payload(message)
        if isinstance(payload, dict):
            if 'method' in payload:
                return cls._process_request(payload)
            else:
                return cls._process_response(payload)
        elif isinstance(payload, list) and cls.allow_batches:
            return cls._process_batch(payload)

        error = payload
        if not isinstance(error, RPCError):
            error = cls.invalid_request('request object must be a dictionary')
        return RPCRequest(error, None, None)

    # Message formation
    @classmethod
    def request_message(cls, item):
        '''Convert an RPCRequest item to a message.'''
        assert isinstance(item, RPCRequest)
        return cls.encode_payload(cls.request_payload(item))

    @classmethod
    def response_message(cls, item):
        '''Convert an RPCResponse item to a message.'''
        assert isinstance(item, RPCResponse)
        return cls.encode_payload(cls.response_payload(item))

    @classmethod
    def error_message(cls, item):
        '''Convert an RPCError item to a message.'''
        assert isinstance(item, RPCError)
        return cls.encode_payload(cls.error_payload(item))

    @classmethod
    def batch_message(cls, item):
        '''Convert an RPCBatch item (a request batch) to a message.'''
        assert isinstance(item, RPCBatch)
        if not cls.allow_batches:
            raise RuntimeError('protocol does not permit batches')
        rb = cls.request_message
        return cls.batch_message_from_parts(rb(req) for req in item)

    @classmethod
    def batch_message_from_parts(cls, messages):
        '''Convert messages, one per batch item, into a batch message.  At
        least one message must be passed.
        '''
        # Comma-separate the messages and wrap the lot in square brackets
        middle = b', '.join(messages)
        assert middle
        return b''.join([b'[', middle, b']'])

    @classmethod
    def encode_payload(cls, payload):
        '''Encode a Python object as JSON and convert it to bytes.'''
        try:
            return json.dumps(payload).encode()
        except TypeError:
            cls.logger.exception(f'payload JSON encoding failure: {payload}')
            if isinstance(payload, dict):
                request_id = payload.get('id')
            else:
                request_id = None
            return cls.error_message(cls.internal_error(request_id))


class JSONRPCv1(JSONRPC):
    '''JSON RPC version 1.0.'''

    allow_batches = False

    @classmethod
    def _message_id(cls, message, require_id):
        # JSONv1 requires an ID always, but without constraint on its type
        # No need to test for a dictionary here as we don't handle batches.
        if 'id' not in message:
            raise cls.invalid_request('request has no "id"')
        return message['id']

    @classmethod
    def _request_args(cls, request):
        args = request.get('params')
        if not isinstance(args, list):
            if args is None:
                raise cls.args_error('no request arguments given')
            raise cls.args_error(f'invalid request arguments: {args}')
        return args

    @classmethod
    def _best_effort_error(cls, error):
        # Do our best to interpret the error
        code = cls.ERROR_CODE_UNAVAILABLE
        message = 'no error message provided'
        if isinstance(error, str):
            message = error
        elif isinstance(error, int):
            code = error
        elif isinstance(error, dict):
            if isinstance(error.get('message'), str):
                message = error['message']
            if isinstance(error.get('code'), int):
                code = error['code']

        return RPCError(code, message)

    @classmethod
    def response_result(cls, payload):
        if 'result' not in payload or 'error' not in payload:
            raise cls.invalid_request(
                'response must contain both "result" and "error"')

        result = payload['result']
        error = payload['error']
        if error is None:
            return result   # It seems None can be a valid result
        if result is not None:
            raise cls.invalid_request(
                'response contains both "result" and "error"')

        return cls._best_effort_error(error)

    @classmethod
    def request_payload(cls, request):
        '''JSON v1 request (or notification) payload.'''
        if isinstance(request.args, dict):
            raise TypeError('JSONRPCv1 does not support named arguments')
        return {
            'method': request.method,
            'params': request.args,
            'id': request.request_id
        }

    @classmethod
    def response_payload(cls, response):
        '''JSON v1 response payload.'''
        result = response.result
        if isinstance(result, RPCError):
            return cls.error_payload(result)
        return {
            'result': result,
            'error': None,
            'id': response.request_id
        }

    @classmethod
    def error_payload(cls, error):
        return {
            'result': None,
            'error': {'code': error.code, 'message': error.message},
            'id': error.request_id
        }


class JSONRPCv2(JSONRPC):
    '''JSON RPC version 2.0.'''

    @classmethod
    def _message_id(cls, message, require_id):
        if not isinstance(message, dict):
            raise cls.invalid_request('request object must be a dictionary')
        if 'id' in message:
            request_id = message['id']
            if not isinstance(request_id, (numbers.Number, str, type(None))):
                raise cls.invalid_request(f'invalid "id": {request_id}')
            return request_id
        else:
            if require_id:
                raise cls.invalid_request('request has no "id"')
            return None

    @classmethod
    def _validate_message(cls, message):
        if message.get('jsonrpc') != '2.0':
            raise cls.invalid_request('"jsonrpc" is not "2.0"')

    @classmethod
    def _request_args(cls, request):
        args = request.get('params', [])
        if not isinstance(args, (dict, list)):
            raise cls.args_error(f'invalid request arguments: {args}')
        return args

    @classmethod
    def response_result(cls, payload):
        if 'result' in payload:
            if 'error' in payload:
                raise cls.invalid_request(
                    'response contains both "result" and "error"')
            return payload['result']

        if 'error' not in payload:
            raise cls.invalid_request(
                'response contains neither "result" nor "error"')

        # Return an RPCError object
        error = payload['error']
        if isinstance(error, dict):
            code = error.get('code')
            message = error.get('message')
            if isinstance(code, int) and isinstance(message, str):
                return RPCError(code, message)

        raise cls.invalid_request(f'ill-formed response error object: {error}')

    @classmethod
    def request_payload(cls, request):
        '''JSON v2 request (or notification) payload.'''
        payload = {
            'jsonrpc': '2.0',
            'method': request.method,
        }
        # A notification?
        if request.request_id is not None:
            payload['id'] = request.request_id
        # Preserve empty dicts as missing params is read as an array
        if request.args or request.args == {}:
            payload['params'] = request.args
        return payload

    @classmethod
    def response_payload(cls, response):
        '''JSON v2 response payload.'''
        result = response.result
        if isinstance(result, RPCError):
            return cls.error_payload(result)

        return {
            'jsonrpc': '2.0',
            'result': result,
            'id': response.request_id
        }

    @classmethod
    def error_payload(cls, error):
        return {
            'jsonrpc': '2.0',
            'error': {'code': error.code, 'message': error.message},
            'id': error.request_id
        }


class JSONRPCLoose(JSONRPC):
    '''A relaxed versin of JSON RPC.'''

    # Don't be so loose we accept any old message ID
    _message_id = JSONRPCv2._message_id
    _validate_message = JSONRPC._validate_message
    _request_args = JSONRPCv2._request_args
    # Outoing messages are JSONRPCv2 so we give the other side the
    # best chance to assume / detect JSONRPCv2 as default protocol.
    error_payload = JSONRPCv2.error_payload
    request_payload = JSONRPCv2.request_payload
    response_payload = JSONRPCv2.response_payload

    @classmethod
    def response_result(cls, payload):
        # Return result, unless it is None and there is an error
        if payload.get('error') is not None:
            if payload.get('result') is not None:
                raise cls.invalid_request(
                    'response contains both "result" and "error"')
            return JSONRPCv1._best_effort_error(payload['error'])

        if 'result' not in payload:
            raise cls.invalid_request(
                'response contains neither "result" nor "error"')

        # Can be None
        return payload['result']


class JSONRPCAutoDetect(JSONRPCv2):

    @classmethod
    def message_to_item(cls, message):
        return cls.detect_protocol(message)

    @classmethod
    def detect_protocol(cls, message):
        '''Attempt to detect the protocol from the message.'''
        main = cls._message_to_payload(message)

        def protocol_for_payload(payload):
            if not isinstance(payload, dict):
                return JSONRPCLoose   # Will error
            # Obey an explicit "jsonrpc"
            version = payload.get('jsonrpc')
            if version == '2.0':
                return JSONRPCv2
            if version == '1.0':
                return JSONRPCv1

            # Now to decide between JSONRPCLoose and JSONRPCv1 if possible
            if 'result' in payload and 'error' in payload:
                return JSONRPCv1
            return JSONRPCLoose

        if isinstance(main, list):
            parts = set(protocol_for_payload(payload) for payload in main)
            # If all same protocol, return it
            if len(parts) == 1:
                return parts.pop()
            # If strict protocol detected, return it, preferring JSONRPCv2.
            # This means a batch of JSONRPCv1 will fail
            for protocol in (JSONRPCv2, JSONRPCv1):
                if protocol in parts:
                    return protocol
            # Will error if no parts
            return JSONRPCLoose

        return protocol_for_payload(main)
